#include "pgkts.h"
#include "ktsdefs.h"
#include "kts.h"
#include "ktsds.h"
#include "ktslib.h"
#include "encoding.h"
#include "debug.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#define GETPATH( x ) x->pathPool[x->sortedIdx[0]]

#define GETTAGINDEX( x, idx ) senP[INFORINDEX(x, idx)].pos - '0'

#define KTS_ENCODING "EUC-KR"

void processXVP( const char* text
	, const char* word
	, PPARSERSTATE state 
	, Path wordPath );

void Startup( void )
{
	char	dict[1024];

	if( getenv( "KDICT_PATH" ) == NULL )
	{
		sprintf( dict, "KDICT_PATH=%s/dict" 
			, getenv( "PWD" ) );

		putenv( dict );
	}

	(void)OpenKTS(0);
}

void Shutdown( void )
{
	(void)CloseKTS(0);
}


void Process( const char* text
	, PPARSERSTATE state )
{
	Trellis*	scanner = NULL;
	Path		wordPath;
	char		enc_text[2048];
	char		dec_text[2048];
	char		token[__MAXMORPH__*2];
	int			word = 0;
	char		hanstr[__MAXMORPH__];
	char		tagstr[10];
	char		eoj_sp;
	int			index = 0;

	memset( enc_text	
		, 0x00
		, sizeof( enc_text ) * sizeof( char ) );

	memset( dec_text	
		, 0x00
		, sizeof( dec_text ) * sizeof( char ) );

	memset( token
		, 0x00
		, sizeof( token ) * sizeof( char ) );

	if( NULL == state->scanner )
	{
		DebugMessage( "Process Scanner = %s\n", text );

		encoding( text
			, KTS_ENCODING
			, state->encoding
			, enc_text );

		(void)MAT( enc_text
			, STATE_BEST
			, 0.0
			, 0.0
			, 0 );

		state->scanner = trellis.nextPtr;
	}
	
	DebugMessage( "PATH IDX = %d\n", state->index );

	scanner = (Trellis* )state->scanner;

	if( scanner == NULL )
	{
		DebugMessage( "Complete 2\n" );
		state->complete = 1;
		return;
	}

	encoding( (const char *)scanner->eojeol
		, state->encoding
		, KTS_ENCODING
		, state->word );
	
	wordPath = GETPATH( scanner );

	GetMorphTag( wordPath
		, state->index
		, hanstr
		, tagstr
		, &eoj_sp );

	index = GETTAGINDEX( wordPath
		, state->index );

	encoding( hanstr
		, state->encoding
		, KTS_ENCODING
		, token );
	
	switch( index )
	{
		case 9: /* 외국어 */
			word = 1;
			state->type = 3;
			break;
		case 12: /* 동작성보통명사 */
			word = 1;
			state->type = 3;
			break;
		case 14: /* 동작성보통명사 */
			word = 1;
			state->type = 3;
			break;
		case 15: /* 동작성보통명사 */
			word = 1;
			state->type = 3;
			break;
		case 18: /* 동작성보통명사 */
			word = 1;
			state->type = 3;
			break;
		case 23: /* 동작성보통명사 */
			word = 1;
			state->type = 3;
			break;
		default:
			state->type = 12;
			word = 0;
			break;		
	}

	state->tlen = strlen( token );
	state->slen += state->tlen;

	if( state->index == PATHLEN( wordPath ) ) 
	{
		state->complete = 1;
		state->index = 0;
		state->slen = 0;
		state->pos += state->tlen;
		state->scanner = NULL;
		DebugMessage( "Complete 3\n" );
	}
	else if( word == 1)
	{
		state->pos += state->tlen;	
		state->complete = 0;
	}
	else
	{		
		if( strlen( text ) == strlen( token ) && 
			PATHLEN(wordPath) + 1 == state->index + 2 &&
			strcmp( text, token ) != 0 )
		{
			state->complete = 1;
			state->index = 0;
			state->slen = 0;
			state->scanner = NULL;
			DebugMessage( "Complete 4\n" );
		}
		else if( index == 50 )
		{
			DebugMessage( "Process XVP\n" );

			processXVP( text
				, token
				, state 
				, wordPath );
		}
		else if( state->hlen > 0 )
		{
			state->tlen = state->hlen;
			state->hlen = 0;
		}
	
		state->pos += state->tlen;
	}		

	state->index ++;

	DebugMessage( "tag = %s, LEN1 =%d, LEN2 = %d word = %s\n"
		, tagstr
		, PATHLEN(wordPath) + 1
		, state->index
		, token );

}

void processXVP( const char* text
	, const char* word
	, PPARSERSTATE state
	, Path wordPath )
{
	int			len = 0;
	char        token[__MAXMORPH__*2];
	char        hanstr[__MAXMORPH__];
	char		*temp = NULL;
	char        tagstr[10];
    char        eoj_sp = ' ';


	if( strncmp( text
		, word
		, state->tlen ) != 0 )
	{
		state->tlen = 0;

		if( state->index + 2 > PATHLEN( wordPath ) + 1 )
			state->hlen = strlen( state->word ) - state->slen;
		else
		{
			GetMorphTag( wordPath
				, state->index + 2
				, hanstr
				, tagstr
				, &eoj_sp );

			encoding( hanstr
				, state->encoding
				, KTS_ENCODING
				, token );

			DebugMessage( "XVP Check Len = %s\n", token );

			temp = strstr( text + state->slen
				, token );			

			len =  (int)(temp - text) - state->slen;

			state->hlen = len;	
		}

		DebugMessage( "XVP Len = %d %d\n", state->hlen
			, state->slen );
	}
	else
	{
		state->hlen = 0;// state->tlen;
	}
}
