#include "pgkts.h"
#include "postgres.h"
#include "utils/builtins.h"
#include "debug.h"
#include "mb/pg_wchar.h"

#include <stdio.h>

typedef struct _LEXDESCR {
	int     lexid;
	char    *alias;
	char    *descr;
} LEXDESCR, *PLEXDESCR;

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

#define NEWPARSERSTATE() (PPARSERSTATE) \
	palloc( sizeof(PARSERSTATE) )

#define GETPARSERSTATE() (PPARSERSTATE)\
	PG_GETARG_POINTER(0)

#define GETLEXDESCR() (PLEXDESCR)\
	palloc( sizeof( LEXDESCR ) * (2+1) )

#define GETTEXT() (char **) PG_GETARG_POINTER(1)

#define GETTEXTLEN() (int *) PG_GETARG_POINTER(2)

#define ISCOMPLETE( s ) ( s->complete == 1 )

#define GETCHAR( s, i ) s->buffer[s->pos + i]

#define ISENDLINE( s ) (s->pos >= s->len)

#define DELETELINE( s ) if( s->line != NULL ) \
	{ free( s->line ); s->line = NULL; }

#define DELETEBUFFER( s ) { pfree( s->buffer ); s->buffer = NULL; }
#define DELETESTATE( s ) if( s != NULL ) \
	{ pfree( s ); s = NULL; }

PG_FUNCTION_INFO_V1(ktsprs_start);
Datum ktsprs_start(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(ktsprs_nexttoken);
Datum ktsprs_nexttoken(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(ktsprs_end);
Datum ktsprs_end(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(ktsprs_lextype);
Datum ktsprs_lextype(PG_FUNCTION_ARGS);


void getWord( PPARSERSTATE state
	, char* word );

void getLine( PPARSERSTATE state );

Datum ktsprs_start( PG_FUNCTION_ARGS )
{
	char			path[1024];
	int				len;
	PPARSERSTATE	state = NEWPARSERSTATE();

	state->len = PG_GETARG_INT32(1);
	len = sizeof( char ) * state->len;

	state->pos = 0;
	state->index = 0;
	state->tlen = 0;
	state->slen = 0;
	state->hlen = 0;
	state->complete = 1;
	state->scanner = NULL;

	state->buffer = (char *)palloc( len );
	
	state->encoding =  (char *)GetDatabaseEncodingName(); 

	strcpy( state->buffer
		, (char *) PG_GETARG_POINTER(0) );

	state->buffer[len] =0x00;

	DebugMessage("------------------------------------\n" );
	DebugMessage("KTS Parser Start\n" );
	DebugMessage( "len = %s, text = %d\n"
		, state->buffer
		, state->len );

	DebugMessage("------------------------------------\n" );

	Startup();
	
	PG_RETURN_POINTER(state);
}

Datum ktsprs_nexttoken( PG_FUNCTION_ARGS )
{
	PPARSERSTATE	state = GETPARSERSTATE();
	char**			t = GETTEXT();
	int*			tlen = GETTEXTLEN();
	char			word[1024];			

	*tlen = state-> pos;
	*t = state->buffer + state->pos;
	
	if( ISENDLINE( state ) == true )
	{
		state->type = 0;
		PG_RETURN_INT32( state->type );
	}

	if( ISCOMPLETE( state ) == true )
	{
		DebugMessage( "Complete 1\n" );
	//	getLine( state );
		state->complete = 0;
		state->slen = 0;
	//	state->index = 0;
	//	state->slen = 0;
	}
	
	switch( GETCHAR( state , 0 ) )
	{
		case '\n':
		case '\r':
		case ' ':
			state->type = 12;				
			state->pos ++;	
			*tlen = 1;
			state->complete = 1;
			state->index = 0;
			state->hlen = 0;
			state->slen = 0;
			DebugMessage( "space \n" );
			break;
		default:
			getWord( state, word );
			Process( word, state );
			*tlen = state->tlen;
			break;
	}


	PG_RETURN_INT32( state->type );
}

Datum ktsprs_lextype( PG_FUNCTION_ARGS )
{
	PLEXDESCR descr = GETLEXDESCR();

	descr[0].lexid = 3;
	descr[0].alias = pstrdup("word");
	descr[0].descr = pstrdup("Word");

	descr[1].lexid = 12;
	descr[1].alias = pstrdup("blank");
	descr[1].descr = pstrdup("Space symbols");

	descr[2].lexid = 0;

	PG_RETURN_POINTER( descr );
}

Datum ktsprs_end( PG_FUNCTION_ARGS )
{
	PPARSERSTATE state = GETPARSERSTATE();

//	DELETELINE( state );
//
	Shutdown();
	DebugMessage("------------------------------------\n" );
	DebugMessage("KTS Parser End\n" );
	DebugMessage("------------------------------------\n" );
	
	DELETEBUFFER( state );

	DELETESTATE( state );

	PG_RETURN_VOID();
}


void getWord( PPARSERSTATE state
	, char* word )
{
	char*	line = NULL;
	char*	token = NULL;
	int		index = 0;
	
	memset( word
		, 0x00
		, sizeof( word ) * sizeof( char ) );

	/*line = state->line;

	for( index = 0; index < strlen( line ); index ++ )
	{
		if( line[index] == ' ' )
		{
			break;
		}
	}

	strncpy( word
		, line
		, index );
*/
	line = state->buffer + state->pos;

	for( index = 0; index < state->len; index ++ )
	{
		if( line[index] == ' ' ||
			line[index] == '\n' ||
			line[index] == '\r' ) 
		{
			break;
		}
	}

	strncpy( word
		, line
		, index );

	word[index] = 0x00;

	DebugMessage( "GetWord len1 = %d, len2 = %d pos = %d, word = %s, line = %s\n"
		, state->len
		, index
		, state->pos
		, word
		, line );
}

void getLine( PPARSERSTATE state )
{

	int		pos = state->pos;
	int		len = state->len - pos;
	int		i = 0;

	DebugMessage( "GL1 pos = %d , len = %d\n", pos, len );

	memset( state->line
		, 0x00
		, sizeof( state->line ) * sizeof( char ) );

	for( i = pos; i < len; i ++ )
	{
		if( state->buffer[i] == '\r' )
		{
			if( state->buffer[i-1] == '\n' ) 
				i -= 2;
			else if( state->buffer[i+1] == '\n' )
				i --;
			else
				i --;
			break;	
		}
		else if( state->buffer[i] == '\r' )
		{
			if( state->buffer[i-1] == '\r' ) 
				i -= 2;
			else if( state->buffer[i+1] == '\r' )
				i --;
			else
				i --;

			break;
		}
	}

	memcpy( state->line 
		, state->buffer + pos
		, i );

	DebugMessage( "Line = %s\n", state->line );
}
